/*
 * (C) Copyright 1996- ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

/// @author Baudouin Raoult
/// @author Tiago Quintino
/// @date   Oct 2016

#ifndef eckit_AutoUmask_h
#define eckit_AutoUmask_h

#include <sys/stat.h>


namespace eckit {


class AutoUmask {
    mode_t umask_;

public:
    explicit AutoUmask(mode_t u = 0) : umask_(::umask(u)) {}
    ~AutoUmask() { ::umask(umask_); }
};


}  // namespace eckit

#endif
