/*
 * (C) Copyright 1996- ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */


#include "eckit/bases/Watcher.h"

//----------------------------------------------------------------------------------------------------------------------

namespace eckit {

//----------------------------------------------------------------------------------------------------------------------


struct DummyWatcher : public Watcher {
    void watch() {}
};

Watcher& Watcher::dummy() {
    static DummyWatcher x;
    return x;
}

//----------------------------------------------------------------------------------------------------------------------

}  // namespace eckit
